@@ -2,14 +2,13 @@
 #include <qpdf/qpdf-config.h>
 
 #include <qpdf/QUtil.hh>
+
 #include <qpdf/PointerHolder.hh>
-#ifdef USE_INSECURE_RANDOM
-# include <qpdf/InsecureRandomDataProvider.hh>
-#endif
-#include <qpdf/SecureRandomDataProvider.hh>
+#include <qpdf/CryptoRandomDataProvider.hh>
 #include <qpdf/QPDFSystemError.hh>
 #include <qpdf/QTC.hh>
 #include <qpdf/QIntC.hh>
+#include <qpdf/Pipeline.hh>
 
 #include <cmath>
 #include <iomanip>
@@ -23,17 +22,36 @@
 #include <stdlib.h>
 #include <string.h>
 #include <fcntl.h>
+#include <memory>
+#include <locale>
+#include <regex>
+#ifndef QPDF_NO_WCHAR_T
+# include <cwchar>
+#endif
 #ifdef _WIN32
-#include <windows.h>
-#include <direct.h>
-#include <io.h>
+# define WIN32_LEAN_AND_MEAN
+# include <windows.h>
+# include <direct.h>
+# include <io.h>
 #else
-#include <unistd.h>
-#include <sys/stat.h>
+# include <unistd.h>
+# include <sys/stat.h>
 #endif
 
-// First element is 128
+// First element is 24
+static unsigned short pdf_doc_low_to_unicode[] = {
+    0x02d8,    // 0x18    BREVE
+    0x02c7,    // 0x19    CARON
+    0x02c6,    // 0x1a    MODIFIER LETTER CIRCUMFLEX ACCENT
+    0x02d9,    // 0x1b    DOT ABOVE
+    0x02dd,    // 0x1c    DOUBLE ACUTE ACCENT
+    0x02db,    // 0x1d    OGONEK
+    0x02da,    // 0x1e    RING ABOVE
+    0x02dc,    // 0x1f    SMALL TILDE
+};
+// First element is 127
 static unsigned short pdf_doc_to_unicode[] = {
+    0xfffd,    // 0x7f    UNDEFINED
     0x2022,    // 0x80    BULLET
     0x2020,    // 0x81    DAGGER
     0x2021,    // 0x82    DOUBLE DAGGER
@@ -234,6 +252,23 @@ static unsigned short mac_roman_to_unicode[] = {
     0x02c7,    // 0xff
 };
 
+class FileCloser
+{
+  public:
+    FileCloser(FILE* f) :
+        f(f)
+    {
+    }
+
+    ~FileCloser()
+    {
+        fclose(f);
+    }
+
+  private:
+    FILE* f;
+};
+
 template <typename T>
 static
 std::string
@@ -248,18 +283,29 @@ int_to_string_base_internal(T num, int base, int length)
         throw std::logic_error(
             "int_to_string_base called with unsupported base");
     }
-    std::ostringstream buf;
-    buf << std::setbase(base) << std::nouppercase << num;
+    std::string cvt;
+    if (base == 10)
+    {
+        // Use the more efficient std::to_string when possible
+        cvt = std::to_string(num);
+    }
+    else
+    {
+        std::ostringstream buf;
+        buf.imbue(std::locale::classic());
+        buf << std::setbase(base) << std::nouppercase << num;
+        cvt = buf.str();
+    }
     std::string result;
-    int str_length = QIntC::to_int(buf.str().length());
+    int str_length = QIntC::to_int(cvt.length());
     if ((length > 0) && (str_length < length))
     {
-	result.append(QIntC::to_size(length - str_length), '0');
+        result.append(QIntC::to_size(length - str_length), '0');
     }
-    result += buf.str();
+    result += cvt;
     if ((length < 0) && (str_length < -length))
     {
-	result.append(QIntC::to_size(-length - str_length), ' ');
+        result.append(QIntC::to_size(-length - str_length), ' ');
     }
     return result;
 }
@@ -290,18 +336,38 @@ QUtil::uint_to_string_base(unsigned long long num, int base, int length)
 
 std::string
 QUtil::double_to_string(double num, int decimal_places)
+{
+    return double_to_string(num, decimal_places, true);
+}
+
+std::string
+QUtil::double_to_string(double num, int decimal_places,
+                        bool trim_trailing_zeroes)
 {
     // Backward compatibility -- this code used to use sprintf and
     // treated decimal_places <= 0 to mean to use the default, which
-    // was six decimal places.  Also sprintf with %*.f interprets the
-    // length as fixed point rather than significant figures.
+    // was six decimal places. Starting in 10.2, we trim trailing
+    // zeroes by default.
     if (decimal_places <= 0)
     {
         decimal_places = 6;
     }
     std::ostringstream buf;
+    buf.imbue(std::locale::classic());
     buf << std::setprecision(decimal_places) << std::fixed << num;
-    return buf.str();
+    std::string result = buf.str();
+    if (trim_trailing_zeroes)
+    {
+        while ((result.length() > 1) && (result.back() == '0'))
+        {
+            result.pop_back();
+        }
+        if ((result.length() > 1) && (result.back() == '.'))
+        {
+            result.pop_back();
+        }
+    }
+    return result;
 }
 
 long long
@@ -389,7 +455,7 @@ QUtil::os_wrapper(std::string const& description, int status)
 {
     if (status == -1)
     {
-	throw_system_error(description);
+        throw_system_error(description);
     }
     return status;
 }
@@ -405,7 +471,7 @@ win_convert_filename(char const* filename)
     size_t len = u16.length();
     size_t wlen = (len / 2) - 1;
     PointerHolder<wchar_t> wfilenamep(true, new wchar_t[wlen + 1]);
-    wchar_t* wfilename = wfilenamep.getPointer();
+    wchar_t* wfilename = wfilenamep.get();
     wfilename[wlen] = 0;
     for (unsigned int i = 2; i < len; i += 2)
     {
@@ -424,9 +490,9 @@ QUtil::safe_fopen(char const* filename, char const* mode)
     FILE* f = 0;
 #ifdef _WIN32
     PointerHolder<wchar_t> wfilenamep = win_convert_filename(filename);
-    wchar_t* wfilename = wfilenamep.getPointer();
+    wchar_t* wfilename = wfilenamep.get();
     PointerHolder<wchar_t> wmodep(true, new wchar_t[strlen(mode) + 1]);
-    wchar_t* wmode = wmodep.getPointer();
+    wchar_t* wmode = wmodep.get();
     wmode[strlen(mode)] = 0;
     for (size_t i = 0; i < strlen(mode); ++i)
     {
@@ -457,11 +523,26 @@ QUtil::fopen_wrapper(std::string const& description, FILE* f)
 {
     if (f == 0)
     {
-	throw_system_error(description);
+        throw_system_error(description);
     }
     return f;
 }
 
+bool
+QUtil::file_can_be_opened(char const* filename)
+{
+    try
+    {
+        fclose(safe_fopen(filename, "rb"));
+        return true;
+    }
+    catch (std::runtime_error&)
+    {
+        // can't open the file
+    }
+    return false;
+}
+
 int
 QUtil::seek(FILE* stream, qpdf_offset_t offset, int whence)
 {
@@ -553,7 +634,7 @@ QUtil::remove_file(char const* path)
 {
 #ifdef _WIN32
     PointerHolder<wchar_t> wpath = win_convert_filename(path);
-    os_wrapper(std::string("remove ") + path, _wunlink(wpath.getPointer()));
+    os_wrapper(std::string("remove ") + path, _wunlink(wpath.get()));
 #else
     os_wrapper(std::string("remove ") + path, unlink(path));
 #endif
@@ -574,13 +655,73 @@ QUtil::rename_file(char const* oldname, char const* newname)
     PointerHolder<wchar_t> wold = win_convert_filename(oldname);
     PointerHolder<wchar_t> wnew = win_convert_filename(newname);
     os_wrapper(std::string("rename ") + oldname + " " + newname,
-               _wrename(wold.getPointer(), wnew.getPointer()));
+               _wrename(wold.get(), wnew.get()));
 #else
     os_wrapper(std::string("rename ") + oldname + " " + newname,
                rename(oldname, newname));
 #endif
 }
 
+void
+QUtil::pipe_file(char const* filename, Pipeline* p)
+{
+    // Exercised in test suite by testing file_provider.
+    FILE* f = safe_fopen(filename, "rb");
+    FileCloser fc(f);
+    size_t len = 0;
+    int constexpr size = 8192;
+    unsigned char buf[size];
+    while ((len = fread(buf, 1, size, f)) > 0)
+    {
+        p->write(buf, len);
+    }
+    p->finish();
+    if (ferror(f))
+    {
+        throw std::runtime_error(
+            std::string("failure reading file ") + filename);
+    }
+}
+
+std::function<void(Pipeline*)>
+QUtil::file_provider(std::string const& filename)
+{
+    return [filename](Pipeline* p) {
+        pipe_file(filename.c_str(), p);
+    };
+}
+
+std::string
+QUtil::path_basename(std::string const& filename)
+{
+#ifdef _WIN32
+    char const* pathsep = "/\\";
+#else
+    char const* pathsep = "/";
+#endif
+    std::string last = filename;
+    auto len = last.length();
+    while (len > 1)
+    {
+        auto pos = last.find_last_of(pathsep);
+        if (pos == len - 1)
+        {
+            last.pop_back();
+            --len;
+        }
+        else if (pos == std::string::npos)
+        {
+            break;
+        }
+        else
+        {
+            last = last.substr(pos + 1);
+            break;
+        }
+    }
+    return last;
+}
+
 char*
 QUtil::copy_string(std::string const& str)
 {
@@ -591,6 +732,26 @@ QUtil::copy_string(std::string const& str)
     return result;
 }
 
+std::shared_ptr<char>
+QUtil::make_shared_cstr(std::string const& str)
+{
+    auto result = QUtil::make_shared_array<char>(str.length() + 1);
+    // Use memcpy in case string contains nulls
+    result.get()[str.length()] = '\0';
+    memcpy(result.get(), str.c_str(), str.length());
+    return result;
+}
+
+std::unique_ptr<char[]>
+QUtil::make_unique_cstr(std::string const& str)
+{
+    auto result = std::make_unique<char[]>(str.length() + 1);
+    // Use memcpy in case string contains nulls
+    result.get()[str.length()] = '\0';
+    memcpy(result.get(), str.c_str(), str.length());
+    return result;
+}
+
 std::string
 QUtil::hex_encode(std::string const& input)
 {
@@ -680,17 +841,17 @@ QUtil::getWhoami(char* argv0)
     if (((whoami = strrchr(argv0, '/')) == NULL) &&
         ((whoami = strrchr(argv0, '\\')) == NULL))
     {
-	whoami = argv0;
+        whoami = argv0;
     }
     else
     {
-	++whoami;
+        ++whoami;
     }
 
     if ((strlen(whoami) > 4) &&
-	(strcmp(whoami + strlen(whoami) - 4, ".exe") == 0))
+        (strcmp(whoami + strlen(whoami) - 4, ".exe") == 0))
     {
-	whoami[strlen(whoami) - 4] = '\0';
+        whoami[strlen(whoami) - 4] = '\0';
     }
 
     return whoami;
@@ -714,9 +875,9 @@ QUtil::get_env(std::string const& var, std::string* value)
 
     if (value)
     {
-	PointerHolder<char> t = PointerHolder<char>(true, new char[len + 1]);
-        ::GetEnvironmentVariable(var.c_str(), t.getPointer(), len);
-	*value = t.getPointer();
+        PointerHolder<char> t = PointerHolder<char>(true, new char[len + 1]);
+        ::GetEnvironmentVariable(var.c_str(), t.get(), len);
+        *value = t.get();
     }
 
     return true;
@@ -762,6 +923,123 @@ QUtil::get_current_time()
 #endif
 }
 
+QUtil::QPDFTime
+QUtil::get_current_qpdf_time()
+{
+#ifdef _WIN32
+    SYSTEMTIME ltime;
+    GetLocalTime(&ltime);
+    TIME_ZONE_INFORMATION tzinfo;
+    GetTimeZoneInformation(&tzinfo);
+    return QPDFTime(static_cast<int>(ltime.wYear),
+                    static_cast<int>(ltime.wMonth),
+                    static_cast<int>(ltime.wDay),
+                    static_cast<int>(ltime.wHour),
+                    static_cast<int>(ltime.wMinute),
+                    static_cast<int>(ltime.wSecond),
+                    // tzinfo.Bias is minutes before UTC
+                    static_cast<int>(tzinfo.Bias));
+#else
+    struct tm ltime;
+    time_t now = time(0);
+    tzset();
+#ifdef HAVE_LOCALTIME_R
+    localtime_r(&now, &ltime);
+#else
+    ltime = *localtime(&now);
+#endif
+#if HAVE_TM_GMTOFF
+    // tm_gmtoff is seconds after UTC
+    int tzoff = -static_cast<int>(ltime.tm_gmtoff / 60);
+#elif HAVE_EXTERN_LONG_TIMEZONE
+    // timezone is seconds before UTC, not adjusted for daylight saving time
+    int tzoff = static_cast<int>(timezone / 60);
+#else
+    // Don't know how to get timezone on this platform
+    int tzoff = 0;
+#endif
+    return QPDFTime(static_cast<int>(ltime.tm_year + 1900),
+                    static_cast<int>(ltime.tm_mon + 1),
+                    static_cast<int>(ltime.tm_mday),
+                    static_cast<int>(ltime.tm_hour),
+                    static_cast<int>(ltime.tm_min),
+                    static_cast<int>(ltime.tm_sec),
+                    tzoff);
+#endif
+}
+
+std::string
+QUtil::qpdf_time_to_pdf_time(QPDFTime const& qtm)
+{
+    std::string tz_offset;
+    int t = qtm.tz_delta;
+    if (t == 0)
+    {
+        tz_offset = "Z";
+    }
+    else
+    {
+        if (t < 0)
+        {
+            t = -t;
+            tz_offset += "+";
+        }
+        else
+        {
+            tz_offset += "-";
+        }
+        tz_offset +=
+            QUtil::int_to_string(t / 60, 2) + "'" +
+            QUtil::int_to_string(t % 60, 2) + "'";
+    }
+    return ("D:" +
+            QUtil::int_to_string(qtm.year, 4) +
+            QUtil::int_to_string(qtm.month, 2) +
+            QUtil::int_to_string(qtm.day, 2) +
+            QUtil::int_to_string(qtm.hour, 2) +
+            QUtil::int_to_string(qtm.minute, 2) +
+            QUtil::int_to_string(qtm.second, 2) +
+            tz_offset);
+}
+
+bool
+QUtil::pdf_time_to_qpdf_time(std::string const& str, QPDFTime* qtm)
+{
+    static std::regex pdf_date("^D:([0-9]{4})([0-9]{2})([0-9]{2})"
+                               "([0-9]{2})([0-9]{2})([0-9]{2})"
+                               "(?:(Z?)|([\\+\\-])([0-9]{2})'([0-9]{2})')$");
+    std::smatch m;
+    if (! std::regex_match(str, m, pdf_date))
+    {
+        return false;
+    }
+    int tz_delta = 0;
+    auto to_i = [](std::string const& s) {
+        return QUtil::string_to_int(s.c_str());
+    };
+
+    if (m[8] != "")
+    {
+        tz_delta = ((to_i(m[9]) * 60) +
+                    to_i(m[10]));
+        if (m[8] == "+")
+        {
+            tz_delta = -tz_delta;
+        }
+    }
+    if (qtm)
+    {
+        *qtm = QPDFTime(to_i(m[1]),
+                        to_i(m[2]),
+                        to_i(m[3]),
+                        to_i(m[4]),
+                        to_i(m[5]),
+                        to_i(m[6]),
+                        tz_delta);
+    }
+    return true;
+}
+
 std::string
 QUtil::toUTF8(unsigned long uval)
 {
@@ -779,42 +1057,42 @@ QUtil::toUTF8(unsigned long uval)
 
     if (uval > 0x7fffffff)
     {
-	throw std::runtime_error("bounds error in QUtil::toUTF8");
+        throw std::runtime_error("bounds error in QUtil::toUTF8");
     }
     else if (uval < 128)
     {
-	result += static_cast<char>(uval);
+        result += static_cast<char>(uval);
     }
     else
     {
-	unsigned char bytes[7];
-	bytes[6] = '\0';
-	unsigned char* cur_byte = &bytes[5];
+        unsigned char bytes[7];
+        bytes[6] = '\0';
+        unsigned char* cur_byte = &bytes[5];
 
-	// maximum value that will fit in the current number of bytes
-	unsigned char maxval = 0x3f; // six bits
+        // maximum value that will fit in the current number of bytes
+        unsigned char maxval = 0x3f; // six bits
 
-	while (uval > maxval)
-	{
-	    // Assign low six bits plus 10000000 to lowest unused
-	    // byte position, then shift
-	    *cur_byte = static_cast<unsigned char>(0x80 + (uval & 0x3f));
-	    uval >>= 6;
-	    // Maximum that will fit in high byte now shrinks by one bit
-	    maxval = static_cast<unsigned char>(maxval >> 1);
-	    // Slide to the left one byte
-	    if (cur_byte <= bytes)
-	    {
-		throw std::logic_error("QUtil::toUTF8: overflow error");
-	    }
-	    --cur_byte;
-	}
-	// If maxval is k bits long, the high (7 - k) bits of the
-	// resulting byte must be high.
-	*cur_byte = static_cast<unsigned char>(
+        while (uval > QIntC::to_ulong(maxval))
+        {
+            // Assign low six bits plus 10000000 to lowest unused
+            // byte position, then shift
+            *cur_byte = static_cast<unsigned char>(0x80 + (uval & 0x3f));
+            uval >>= 6;
+            // Maximum that will fit in high byte now shrinks by one bit
+            maxval = static_cast<unsigned char>(maxval >> 1);
+            // Slide to the left one byte
+            if (cur_byte <= bytes)
+            {
+                throw std::logic_error("QUtil::toUTF8: overflow error");
+            }
+            --cur_byte;
+        }
+        // If maxval is k bits long, the high (7 - k) bits of the
+        // resulting byte must be high.
+        *cur_byte = static_cast<unsigned char>(
             QIntC::to_ulong(0xff - (1 + (maxval << 1))) + uval);
 
-	result += reinterpret_cast<char*>(cur_byte);
+        result += reinterpret_cast<char*>(cur_byte);
     }
 
     return result;
@@ -859,6 +1137,63 @@ QUtil::toUTF16(unsigned long uval)
 
 // Random data support
 
+class RandomDataProviderProvider
+{
+  public:
+    RandomDataProviderProvider();
+    void setProvider(RandomDataProvider*);
+    RandomDataProvider* getProvider();
+
+  private:
+    RandomDataProvider* default_provider;
+    RandomDataProvider* current_provider;
+};
+
+RandomDataProviderProvider::RandomDataProviderProvider() :
+    default_provider(CryptoRandomDataProvider::getInstance()),
+    current_provider(0)
+{
+    this->current_provider = default_provider;
+}
+
+RandomDataProvider*
+RandomDataProviderProvider::getProvider()
+{
+    return this->current_provider;
+}
+
+void
+RandomDataProviderProvider::setProvider(RandomDataProvider* p)
+{
+    this->current_provider = p ? p : this->default_provider;
+}
+
+static RandomDataProviderProvider*
+getRandomDataProviderProvider()
+{
+    // Thread-safe static initializer
+    static RandomDataProviderProvider rdpp;
+    return &rdpp;
+}
+
+void
+QUtil::setRandomDataProvider(RandomDataProvider* p)
+{
+    getRandomDataProviderProvider()->setProvider(p);
+}
+
+RandomDataProvider*
+QUtil::getRandomDataProvider()
+{
+    return getRandomDataProviderProvider()->getProvider();
+}
+
+void
+QUtil::initializeWithRandomBytes(unsigned char* data, size_t len)
+{
+    getRandomDataProvider()->provideRandomData(data, len);
+}
+
 long
 QUtil::random()
 {
@@ -869,69 +1204,6 @@ QUtil::random()
     return result;
 }
 
-static RandomDataProvider* random_data_provider = 0;
-
-#ifdef USE_INSECURE_RANDOM
-static RandomDataProvider* insecure_random_data_provider =
-    InsecureRandomDataProvider::getInstance();
-#else
-static RandomDataProvider* insecure_random_data_provider = 0;
-#endif
-static RandomDataProvider* secure_random_data_provider =
-    SecureRandomDataProvider::getInstance();
-
-static void
-initialize_random_data_provider()
-{
-    if (random_data_provider == 0)
-    {
-        if (secure_random_data_provider)
-        {
-            random_data_provider = secure_random_data_provider;
-        }
-        else if (insecure_random_data_provider)
-        {
-            random_data_provider = insecure_random_data_provider;
-        }
-    }
-    // QUtil.hh has comments indicating that getRandomDataProvider(),
-    // which calls this method, never returns null.
-    if (random_data_provider == 0)
-    {
-        throw std::logic_error("QPDF has no random data provider");
-    }
-}
-
-void
-QUtil::setRandomDataProvider(RandomDataProvider* p)
-{
-    random_data_provider = p;
-}
-
-RandomDataProvider*
-QUtil::getRandomDataProvider()
-{
-    initialize_random_data_provider();
-    return random_data_provider;
-}
-
-void
-QUtil::initializeWithRandomBytes(unsigned char* data, size_t len)
-{
-    initialize_random_data_provider();
-    random_data_provider->provideRandomData(data, len);
-}
-
-void
-QUtil::srandom(unsigned int seed)
-{
-#ifdef HAVE_RANDOM
-    ::srandom(seed);
-#else
-    srand(seed);
-#endif
-}
-
 bool
 QUtil::is_hex_digit(char ch)
 {
@@ -987,30 +1259,18 @@ QUtil::is_number(char const* p)
     return found_digit;
 }
 
-std::list<std::string>
-QUtil::read_lines_from_file(char const* filename)
-{
-    std::ifstream in(filename, std::ios_base::binary);
-    if (! in.is_open())
-    {
-        throw_system_error(std::string("open ") + filename);
-    }
-    std::list<std::string> lines = read_lines_from_file(in);
-    in.close();
-    return lines;
-}
-
 void
 QUtil::read_file_into_memory(
     char const* filename,
     PointerHolder<char>& file_buf, size_t& size)
 {
     FILE* f = safe_fopen(filename, "rb");
+    FileCloser fc(f);
     fseek(f, 0, SEEK_END);
     size = QIntC::to_size(QUtil::tell(f));
     fseek(f, 0, SEEK_SET);
-    file_buf = PointerHolder<char>(true, new char[size]);
-    char* buf_p = file_buf.getPointer();
+    file_buf = make_array_pointer_holder<char>(size);
+    char* buf_p = file_buf.get();
     size_t bytes_read = 0;
     size_t len = 0;
     while ((len = fread(buf_p + bytes_read, 1, size - bytes_read, f)) > 0)
@@ -1036,46 +1296,93 @@ QUtil::read_file_into_memory(
                 uint_to_string(size));
         }
     }
-    fclose(f);
+}
+
+static bool read_char_from_FILE(char& ch, FILE* f)
+{
+    auto len = fread(&ch, 1, 1, f);
+    if (len == 0)
+    {
+        if (ferror(f))
+        {
+            throw std::runtime_error("failure reading character from file");
+        }
+        return false;
+    }
+    return true;
 }
 
 std::list<std::string>
-QUtil::read_lines_from_file(std::istream& in)
+QUtil::read_lines_from_file(char const* filename, bool preserve_eol)
+{
+    std::list<std::string> lines;
+    FILE* f = safe_fopen(filename, "rb");
+    FileCloser fc(f);
+    auto next_char = [&f](char& ch) { return read_char_from_FILE(ch, f); };
+    read_lines_from_file(next_char, lines, preserve_eol);
+    return lines;
+}
+
+std::list<std::string>
+QUtil::read_lines_from_file(std::istream& in, bool preserve_eol)
+{
+    std::list<std::string> lines;
+    auto next_char = [&in](char& ch) { return (in.get(ch)) ? true: false; };
+    read_lines_from_file(next_char, lines, preserve_eol);
+    return lines;
+}
+
+std::list<std::string>
+QUtil::read_lines_from_file(FILE* f, bool preserve_eol)
+{
+    std::list<std::string> lines;
+    auto next_char = [&f](char& ch) { return read_char_from_FILE(ch, f); };
+    read_lines_from_file(next_char, lines, preserve_eol);
+    return lines;
+}
+
+void
+QUtil::read_lines_from_file(std::function<bool(char&)> next_char,
+                            std::list<std::string>& lines,
+                            bool preserve_eol)
 {
-    std::list<std::string> result;
     std::string* buf = 0;
-
     char c;
-    while (in.get(c))
+    while (next_char(c))
     {
-	if (buf == 0)
-	{
-	    result.push_back("");
-	    buf = &(result.back());
-	    buf->reserve(80);
-	}
+        if (buf == 0)
+        {
+            lines.push_back("");
+            buf = &(lines.back());
+            buf->reserve(80);
+        }
 
-	if (buf->capacity() == buf->size())
-	{
-	    buf->reserve(buf->capacity() * 2);
-	}
-	if (c == '\n')
-	{
-            // Remove any carriage return that preceded the
-            // newline and discard the newline
-            if ((! buf->empty()) && ((*(buf->rbegin())) == '\r'))
+        if (buf->capacity() == buf->size())
+        {
+            buf->reserve(buf->capacity() * 2);
+        }
+        if (c == '\n')
+        {
+            if (preserve_eol)
             {
-                buf->erase(buf->length() - 1);
+                buf->append(1, c);
             }
-	    buf = 0;
-	}
-	else
-	{
-	    buf->append(1, c);
-	}
+            else
+            {
+                // Remove any carriage return that preceded the
+                // newline and discard the newline
+                if ((! buf->empty()) && ((*(buf->rbegin())) == '\r'))
+                {
+                    buf->erase(buf->length() - 1);
+                }
+            }
+            buf = 0;
+        }
+        else
+        {
+            buf->append(1, c);
+        }
     }
-
-    return result;
 }
 
 int
@@ -1758,6 +2065,30 @@ encode_pdfdoc(unsigned long codepoint)
     unsigned char ch = '\0';
     switch (codepoint)
     {
+      case 0x02d8:
+        ch = 0x18;
+        break;
+      case 0x02c7:
+        ch = 0x19;
+        break;
+      case 0x02c6:
+        ch = 0x1a;
+        break;
+      case 0x02d9:
+        ch = 0x1b;
+        break;
+      case 0x02dd:
+        ch = 0x1c;
+        break;
+      case 0x02db:
+        ch = 0x1d;
+        break;
+      case 0x02da:
+        ch = 0x1e;
+        break;
+      case 0x02dc:
+        ch = 0x1f;
+        break;
       case 0x2022:
         ch = 0x80;
         break;
@@ -2116,12 +2447,12 @@ QUtil::win_ansi_to_utf8(std::string const& val)
     for (unsigned int i = 0; i < len; ++i)
     {
         unsigned char ch = static_cast<unsigned char>(val.at(i));
-        unsigned short val = ch;
+        unsigned short ch_short = ch;
         if ((ch >= 128) && (ch <= 160))
         {
-            val = win_ansi_to_unicode[ch - 128];
+            ch_short = win_ansi_to_unicode[ch - 128];
         }
-        result += QUtil::toUTF8(val);
+        result += QUtil::toUTF8(ch_short);
     }
     return result;
 }
@@ -2134,12 +2465,12 @@ QUtil::mac_roman_to_utf8(std::string const& val)
     for (unsigned int i = 0; i < len; ++i)
     {
         unsigned char ch = static_cast<unsigned char>(val.at(i));
-        unsigned short val = ch;
+        unsigned short ch_short = ch;
         if (ch >= 128)
         {
-            val = mac_roman_to_unicode[ch - 128];
+            ch_short = mac_roman_to_unicode[ch - 128];
         }
-        result += QUtil::toUTF8(val);
+        result += QUtil::toUTF8(ch_short);
     }
     return result;
 }
@@ -2152,12 +2483,20 @@ QUtil::pdf_doc_to_utf8(std::string const& val)
     for (unsigned int i = 0; i < len; ++i)
     {
         unsigned char ch = static_cast<unsigned char>(val.at(i));
-        unsigned short val = ch;
-        if ((ch >= 128) && (ch <= 160))
+        unsigned short ch_short = ch;
+        if ((ch >= 127) && (ch <= 160))
         {
-            val = pdf_doc_to_unicode[ch - 128];
+            ch_short = pdf_doc_to_unicode[ch - 127];
         }
-        result += QUtil::toUTF8(val);
+        else if ((ch >= 24) && (ch <= 31))
+        {
+            ch_short = pdf_doc_low_to_unicode[ch - 24];
+        }
+        else if (ch == 173)
+        {
+            ch_short = 0xfffd;
+        }
+        result += QUtil::toUTF8(ch_short);
     }
     return result;
 }
@@ -2287,3 +2626,59 @@ QUtil::possible_repaired_encodings(std::string supplied)
     }
     return t;
 }
+
+#ifndef QPDF_NO_WCHAR_T
+static int
+call_main_from_wmain(bool, int argc, wchar_t const* const argv[],
+                     std::function<int(int, char*[])> realmain)
+{
+    // argv contains UTF-16-encoded strings with a 16-bit wchar_t.
+    // Convert this to UTF-8-encoded strings for compatibility with
+    // other systems. That way the rest of qpdf.cc can just act like
+    // arguments are UTF-8.
+
+    std::vector<std::unique_ptr<char[]>> utf8_argv;
+    for (int i = 0; i < argc; ++i)
+    {
+        std::string utf16;
+        for (size_t j = 0; j < std::wcslen(argv[i]); ++j)
+        {
+            unsigned short codepoint = static_cast<unsigned short>(argv[i][j]);
+            utf16.append(1, static_cast<char>(
+                             QIntC::to_uchar(codepoint >> 8)));
+            utf16.append(1, static_cast<char>(
+                             QIntC::to_uchar(codepoint & 0xff)));
+        }
+        std::string utf8 = QUtil::utf16_to_utf8(utf16);
+        utf8_argv.push_back(QUtil::make_unique_cstr(utf8));
+    }
+    auto utf8_argv_sp = std::make_unique<char*[]>(1+utf8_argv.size());
+    char** new_argv = utf8_argv_sp.get();
+    for (size_t i = 0; i < utf8_argv.size(); ++i)
+    {
+        new_argv[i] = utf8_argv.at(i).get();
+    }
+    argc = QIntC::to_int(utf8_argv.size());
+    new_argv[argc] = 0;
+    return realmain(argc, new_argv);
+}
+
+int
+QUtil::call_main_from_wmain(int argc, wchar_t* argv[],
+                            std::function<int(int, char*[])> realmain)
+{
+    return ::call_main_from_wmain(true, argc, argv, realmain);
+}
+
+int
+QUtil::call_main_from_wmain(
+    int argc, wchar_t const* const argv[],
+    std::function<int(int, char const* const[])> realmain)
+{
+    return ::call_main_from_wmain(
+        true, argc, argv, [realmain](int new_argc, char* new_argv[]) {
+            return realmain(new_argc, new_argv);
+        });
+}
+
+#endif // QPDF_NO_WCHAR_T
